/**
 *  psd.ivp
 */
#include "stdafx.h"
#include "../ivp.h"

#include "psd.hxx"
#include "rle.hxx"

inline static WORD bswap(WORD x) { return _byteswap_ushort(x); }
inline static DWORD bswap(DWORD x) { return _byteswap_ulong(x); }

template <typename T>
inline static T * Alloc(SIZE_T n)
{
	return static_cast<T *>(::HeapAlloc(::GetProcessHeap(), 0, sizeof(T) * n));
}
inline static void Free(LPVOID p)
{
	::HeapFree(::GetProcessHeap(), 0, p);
}

class ScanlineReader : public IByteReader
{
public:
	ScanlineReader(LPBYTE ptr, DWORD width)
		: Ptr(ptr), End(ptr + width)
	{
	}
	BOOL Get(BYTE &x)
	{
		if (Ptr >= End)
			return FALSE;
		x = *Ptr++;
		return TRUE;
	}
private:
	LPBYTE Ptr;
	LPBYTE End;
};

class ChannelWriter : public IByteWriter
{
public:
	ChannelWriter(LPBYTE ptr, DWORD channels, DWORD width)
		: Ptr(ptr), End(Ptr + channels * width), Channels(channels)
	{
	}
	BOOL Put(BYTE x)
	{
		if (Ptr >= End)
			return FALSE;
		*Ptr = x;
		Ptr += Channels;
		return TRUE;
	}
private:
	LPBYTE Ptr;
	LPBYTE End;
	DWORD  Channels;
};

inline static void CMYKtoRGB(LPBYTE p)
{
	BYTE c = p[0], m = p[1], y = p[2], k = p[3];
	p[0] = static_cast<BYTE>(k - ((0xFF - y) * k * 0x101 + 0x8080) / 0x10000);
	p[1] = static_cast<BYTE>(k - ((0xFF - m) * k * 0x101 + 0x8080) / 0x10000);
	p[2] = static_cast<BYTE>(k - ((0xFF - c) * k * 0x101 + 0x8080) / 0x10000);
	p[3] = 0xFF;
}

static LPCWSTR Patterns[] = { L"*.psd", NULL };

static HBITMAP GetImage(IStream *Stream)
{
	PSD::FileHeader header;
	if (Stream->Read(&header, sizeof header, NULL) != S_OK)
		return NULL;
	if (header.Signature != PSD::FILE_SIGNATURE || bswap(header.Version) != PSD::FILE_VERSION)
		return NULL;
	if (header.Reserved[0] != 0 || header.Reserved[1] != 0 || header.Reserved[2] != 0)
		return NULL;
	PSD::ColorModeData color;
	if (Stream->Read(&color, sizeof color, NULL) != S_OK)
		return NULL;
	WORD mode = bswap(header.Mode);
	WORD channels = bswap(header.Channels);
	LONG height = static_cast<LONG>(bswap(header.Height));
	LONG width = static_cast<LONG>(bswap(header.Width));
	WORD depth = bswap(header.Depth);
	if (width <= 0 || height <= 0 || depth != 8)
		return NULL;
	BYTE bmibuf[sizeof(BITMAPINFOHEADER) + sizeof(RGBQUAD) * 0x100];
	::ZeroMemory(bmibuf, sizeof bmibuf);
	BITMAPINFO &bmi = reinterpret_cast<BITMAPINFO &>(bmibuf);
	bmi.bmiHeader.biSize   = sizeof bmi.bmiHeader;
	bmi.bmiHeader.biWidth  = width;
	bmi.bmiHeader.biHeight = height;
	bmi.bmiHeader.biPlanes = 1;
	switch (mode) {
	case PSD::MODE_GRAYSCALE:
		if (channels != 1 || color.Length != 0)
			return NULL;
		bmi.bmiHeader.biBitCount = 8;
		for (UINT i = 0; i < 0x100; i++)
			reinterpret_cast<COLORREF &>(bmi.bmiColors[i]) = RGB(i, i, i);
		break;
	case PSD::MODE_INDEXED:
		if (channels != 1 || bswap(color.Length) != 0x300)
			return NULL;
		bmi.bmiHeader.biBitCount = 8;
		for (WORD ch = 0; ch < 3; ch++) {
			BYTE palette[0x100];
			if (Stream->Read(palette, sizeof palette, NULL) != S_OK)
				return NULL;
			for (UINT i = 0; i < 0x100; i++)
				reinterpret_cast<LPBYTE>(&bmi.bmiColors[i])[2 - ch] = palette[i];
		}
		break;
	case PSD::MODE_RGB:
		if (channels != 3 && channels != 4 || color.Length != 0)
			return NULL;
		bmi.bmiHeader.biBitCount = static_cast<WORD>(8 * channels);
		break;
	case PSD::MODE_CMYK:
		if (channels != 4 || color.Length != 0)
			return NULL;
		bmi.bmiHeader.biBitCount = 32;
		break;
	default:
		// not supported
		return NULL;
	}
	
	// skip metadata
	PSD::ImageResources resources;
	if (Stream->Read(&resources, sizeof resources, NULL) != S_OK)
		return NULL;
	LARGE_INTEGER offset;
	offset.QuadPart = bswap(resources.Length);
	if (Stream->Seek(offset, STREAM_SEEK_CUR, NULL) != S_OK)
		return NULL;
	PSD::LayerAndMaskInfo layer;
	if (Stream->Read(&layer, sizeof layer, NULL) != S_OK)
		return NULL;
	offset.QuadPart = bswap(layer.Length);
	if (Stream->Seek(offset, STREAM_SEEK_CUR, NULL) != S_OK)
		return NULL;
	
	// read image data
	PSD::ImageData data;
	if (Stream->Read(&data, sizeof data, NULL) != S_OK)
		return NULL;
	WORD compression = bswap(data.Compression);
	if (compression != PSD::COMPRESSION_RAW && compression != PSD::COMPRESSION_RLE)
		return NULL;
	LPVOID pBits = NULL;
	HBITMAP hBmp = ::CreateDIBSection(NULL, &bmi, DIB_RGB_COLORS, &pBits, NULL, 0);
	if (!hBmp)
		return NULL;
	LPBYTE ptr = static_cast<LPBYTE>(pBits);
	DWORD stride = (bmi.bmiHeader.biWidth * bmi.bmiHeader.biBitCount + 31) >> 3 & ~3;
	WORD chmap[4] = { 0, 1, 2, 3 };
	if (mode == PSD::MODE_RGB)
		chmap[0] = 2, chmap[2] = 0;
	if (compression == PSD::COMPRESSION_RAW) {
		LPBYTE buffer = Alloc<BYTE>(width);
		if (!buffer)
			goto error;
		for (WORD ch = 0; ch < channels; ch++) {
			for (LONG y = 0; y < height; y++) {
				ULONG length = 0;
				Stream->Read(buffer, width, &length);
				ScanlineReader reader(buffer, length);
				ChannelWriter writer(ptr + stride * (height - 1 - y) + chmap[ch], channels, width);
				for (BYTE x; reader.Get(x); )
					writer.Put(x);
			}
		}
		Free(buffer);
	} else {
		WORD *limits = Alloc<WORD>(channels * height);
		if (Stream->Read(limits, sizeof(WORD) * channels * height, NULL) != S_OK) {
			Free(limits);
			goto error;
		}
		DWORD maxlimit = width;
		for (DWORD i = 0; i < channels * static_cast<DWORD>(height); i++)
			maxlimit = max(maxlimit, limits[i]);
		LPBYTE buffer = Alloc<BYTE>(maxlimit);
		if (!buffer) {
			Free(limits);
			goto error;
		}
		for (WORD ch = 0; ch < channels; ch++) {
			for (LONG y = 0; y < height; y++) {
				ULONG length = bswap(limits[height * ch + y]);
				Stream->Read(buffer, length, &length);
				ScanlineReader reader(buffer, length);
				ChannelWriter writer(ptr + stride * (height - 1 - y) + chmap[ch], channels, width);
				UnpackBits(&reader, &writer);
			}
		}
		Free(buffer);
		Free(limits);
	}
	if (mode == PSD::MODE_CMYK) {
		for (LONG y = 0; y < height; y++)
			for (LONG x = 0; x < width; x++)
				CMYKtoRGB(ptr + stride * (height - 1 - y) + 4 * x);
	}
	return hBmp;
	
error:
	::DeleteObject(hBmp);
	return NULL;
}

IVP_API BOOL WINAPI GetIVP1(IVP1 *IVP)
{
	IVP->Patterns = Patterns;
	IVP->GetImage = GetImage;
	return TRUE;
}
